/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { Trace } from "@fluidframework/common-utils";
import * as Base from "./base";
import * as Collections from "./collections";
import { LocalReference, LocalReferenceCollection } from "./localReference";
import { IMergeTreeDeltaOpArgs, MergeTreeDeltaCallback, MergeTreeMaintenanceCallback } from "./mergeTreeDeltaCallback";
import { TrackingGroupCollection } from "./mergeTreeTracking";
import * as ops from "./ops";
import { PartialSequenceLengths } from "./partialLengths";
import * as Properties from "./properties";
import { SegmentGroupCollection } from "./segmentGroupCollection";
import { PropertiesManager } from "./segmentPropertiesManager";
export interface ReferencePosition {
    properties?: Properties.PropertySet;
    refType: ops.ReferenceType;
    isLeaf(): boolean;
    getSegment(): ISegment | undefined;
    getOffset(): number;
    addProperties(newProps: Properties.PropertySet, op?: ops.ICombiningOp): any;
    hasTileLabels(): boolean;
    hasRangeLabels(): boolean;
    hasTileLabel(label: string): boolean;
    hasRangeLabel(label: string): boolean;
    getTileLabels(): string[] | undefined;
    getRangeLabels(): string[] | undefined;
}
export declare type RangeStackMap = Properties.MapLike<Collections.Stack<ReferencePosition>>;
export interface IMergeNodeCommon {
    parent?: IMergeBlock;
    /**
     * The length of the contents of the node.
     */
    cachedLength: number;
    index: number;
    ordinal: string;
    isLeaf(): this is ISegment;
}
export declare type IMergeNode = IMergeBlock | ISegment;
export interface IMergeBlock extends IMergeNodeCommon {
    needsScour?: boolean;
    childCount: number;
    children: IMergeNode[];
    partialLengths?: PartialSequenceLengths;
    hierBlock(): IHierBlock | undefined;
    assignChild(child: IMergeNode, index: number, updateOrdinal?: boolean): any;
    setOrdinal(child: IMergeNode, index: number): any;
}
export interface IHierBlock extends IMergeBlock {
    hierToString(indentCount: number): any;
    addNodeReferences(mergeTree: MergeTree, node: IMergeNode): any;
    rightmostTiles: Properties.MapLike<ReferencePosition>;
    leftmostTiles: Properties.MapLike<ReferencePosition>;
    rangeStacks: RangeStackMap;
}
export interface IRemovalInfo {
    removedSeq?: number;
    removedClientId?: number;
    removedClientOverlap?: number[];
}
/**
 * A segment representing a portion of the merge tree.
 */
export interface ISegment extends IMergeNodeCommon, IRemovalInfo {
    readonly type: string;
    readonly segmentGroups: SegmentGroupCollection;
    readonly trackingCollection: TrackingGroupCollection;
    propertyManager?: PropertiesManager;
    localSeq?: number;
    localRemovedSeq?: number;
    seq?: number;
    clientId: number;
    localRefs?: LocalReferenceCollection;
    properties?: Properties.PropertySet;
    addProperties(newProps: Properties.PropertySet, op?: ops.ICombiningOp, seq?: number, collabWindow?: CollaborationWindow): Properties.PropertySet | undefined;
    clone(): ISegment;
    canAppend(segment: ISegment): boolean;
    append(segment: ISegment): void;
    splitAt(pos: number): ISegment | undefined;
    toJSONObject(): any;
    /**
     * Acks the current segment against the segment group, op, and merge tree.
     *
     * Throws error if the segment state doesn't match segment group or op.
     * E.g. Segment group not first is pending queue.
     * Inserted segment does not have unassigned sequence number.
     *
     * Returns true if the op  modifies the segment, otherwise false.
     * The only current false case is overlapping remove, where a segment is removed
     * by a previously sequenced operation before the current operation is acked.
     */
    ack(segmentGroup: SegmentGroup, opArgs: IMergeTreeDeltaOpArgs, mergeTree: MergeTree): boolean;
}
export interface IMarkerModifiedAction {
    (marker: Marker): void;
}
export interface ISegmentAction<TClientData> {
    (segment: ISegment, pos: number, refSeq: number, clientId: number, start: number, end: number, accum: TClientData): boolean;
}
export interface ISegmentChanges {
    next?: ISegment;
    replaceCurrent?: ISegment;
}
export interface BlockAction<TClientData> {
    (block: IMergeBlock, pos: number, refSeq: number, clientId: number, start: number | undefined, end: number | undefined, accum: TClientData): boolean;
}
export interface NodeAction<TClientData> {
    (node: IMergeNode, pos: number, refSeq: number, clientId: number, start: number | undefined, end: number | undefined, clientData: TClientData): boolean;
}
export interface IncrementalSegmentAction<TContext> {
    (segment: ISegment, state: IncrementalMapState<TContext>): any;
}
export interface IncrementalBlockAction<TContext> {
    (state: IncrementalMapState<TContext>): any;
}
export interface BlockUpdateActions {
    child: (block: IMergeBlock, index: number) => void;
}
export interface InsertContext {
    candidateSegment?: ISegment;
    prepareEvents?: boolean;
    structureChange?: boolean;
    leaf: (segment: ISegment | undefined, pos: number, ic: InsertContext) => ISegmentChanges;
    continuePredicate?: (continueFromBlock: IMergeBlock) => boolean;
}
export interface SegmentActions<TClientData> {
    leaf?: ISegmentAction<TClientData>;
    shift?: NodeAction<TClientData>;
    contains?: NodeAction<TClientData>;
    pre?: BlockAction<TClientData>;
    post?: BlockAction<TClientData>;
}
export interface IncrementalSegmentActions<TContext> {
    leaf: IncrementalSegmentAction<TContext>;
    pre?: IncrementalBlockAction<TContext>;
    post?: IncrementalBlockAction<TContext>;
}
export interface SearchResult {
    text: string;
    pos: number;
}
export interface MergeTreeStats {
    maxHeight: number;
    nodeCount: number;
    leafCount: number;
    removedLeafCount: number;
    liveCount: number;
    histo: number[];
    windowTime?: number;
    packTime?: number;
    ordTime?: number;
    maxOrdTime?: number;
}
export interface SegmentGroup {
    segments: ISegment[];
    localSeq: number;
}
export declare class MergeNode implements IMergeNodeCommon {
    index: number;
    ordinal: string;
    parent?: IMergeBlock;
    cachedLength: number;
    isLeaf(): boolean;
}
export declare function ordinalToArray(ord: string): number[];
export declare const MaxNodesInBlock = 8;
export declare class MergeBlock extends MergeNode implements IMergeBlock {
    childCount: number;
    static traceOrdinals: boolean;
    children: IMergeNode[];
    constructor(childCount: number);
    hierBlock(): HierMergeBlock | undefined;
    setOrdinal(child: IMergeNode, index: number): void;
    assignChild(child: IMergeNode, index: number, updateOrdinal?: boolean): void;
}
declare class HierMergeBlock extends MergeBlock implements IMergeBlock {
    rightmostTiles: Properties.MapLike<ReferencePosition>;
    leftmostTiles: Properties.MapLike<ReferencePosition>;
    rangeStacks: Properties.MapLike<Collections.Stack<ReferencePosition>>;
    constructor(childCount: number);
    addNodeReferences(mergeTree: MergeTree, node: IMergeNode): void;
    hierBlock(): this;
    hierToString(indentCount: number): string;
}
export declare abstract class BaseSegment extends MergeNode implements ISegment {
    constructor();
    clientId: number;
    seq: number;
    removedSeq?: number;
    removedClientId?: number;
    removedClientOverlap?: number[];
    readonly segmentGroups: SegmentGroupCollection;
    readonly trackingCollection: TrackingGroupCollection;
    propertyManager?: PropertiesManager;
    properties?: Properties.PropertySet;
    localRefs?: LocalReferenceCollection;
    abstract readonly type: string;
    localSeq?: number;
    localRemovedSeq?: number;
    addProperties(newProps: Properties.PropertySet, op?: ops.ICombiningOp, seq?: number, collabWindow?: CollaborationWindow): Properties.PropertySet | undefined;
    hasProperty(key: string): boolean;
    isLeaf(): boolean;
    cloneInto(b: ISegment): void;
    canAppend(segment: ISegment): boolean;
    addSerializedProps(jseg: ops.IJSONSegment): void;
    abstract toJSONObject(): any;
    ack(segmentGroup: SegmentGroup, opArgs: IMergeTreeDeltaOpArgs, mergeTree: MergeTree): boolean;
    splitAt(pos: number): ISegment | undefined;
    private copyPropertiesTo;
    abstract clone(): ISegment;
    abstract append(segment: ISegment): void;
    protected abstract createSplitSegmentAt(pos: number): BaseSegment | undefined;
}
export declare const reservedTileLabelsKey = "referenceTileLabels";
export declare const reservedRangeLabelsKey = "referenceRangeLabels";
export declare const reservedMarkerIdKey = "markerId";
export declare const reservedMarkerSimpleTypeKey = "markerSimpleType";
export declare const refGetTileLabels: (refPos: ReferencePosition) => string[] | undefined;
export declare const refGetRangeLabels: (refPos: ReferencePosition) => string[] | undefined;
export declare function refHasTileLabel(refPos: ReferencePosition, label: string): boolean;
export declare function refHasRangeLabel(refPos: ReferencePosition, label: string): boolean;
export interface IJSONMarkerSegment extends ops.IJSONSegment {
    marker: ops.IMarkerDef;
}
export declare class Marker extends BaseSegment implements ReferencePosition {
    refType: ops.ReferenceType;
    static readonly type = "Marker";
    static is(segment: ISegment): segment is Marker;
    readonly type = "Marker";
    static make(refType: ops.ReferenceType, props?: Properties.PropertySet): Marker;
    constructor(refType: ops.ReferenceType);
    toJSONObject(): IJSONMarkerSegment;
    static fromJSONObject(spec: any): Marker | undefined;
    clone(): Marker;
    getSegment(): this;
    getOffset(): number;
    hasSimpleType(simpleTypeName: string): boolean;
    getProperties(): Properties.PropertySet | undefined;
    getId(): string | undefined;
    hasTileLabels(): boolean;
    hasRangeLabels(): any;
    hasTileLabel(label: string): boolean;
    hasRangeLabel(label: string): boolean;
    getTileLabels(): any;
    getRangeLabels(): any;
    toString(): string;
    protected createSplitSegmentAt(pos: number): undefined;
    canAppend(segment: ISegment): boolean;
    append(): void;
}
export declare enum IncrementalExecOp {
    Go = 0,
    Stop = 1,
    Yield = 2
}
export declare class IncrementalMapState<TContext> {
    block: IMergeBlock;
    actions: IncrementalSegmentActions<TContext>;
    pos: number;
    refSeq: number;
    clientId: number;
    context: TContext;
    start: number;
    end: number;
    childIndex: number;
    op: IncrementalExecOp;
    constructor(block: IMergeBlock, actions: IncrementalSegmentActions<TContext>, pos: number, refSeq: number, clientId: number, context: TContext, start: number, end: number, childIndex?: number);
}
export declare class CollaborationWindow {
    clientId: number;
    collaborating: boolean;
    minSeq: number;
    currentSeq: number;
    localSeq: number;
    loadFrom(a: CollaborationWindow): void;
}
export declare const compareNumbers: (a: number, b: number) => number;
export declare const compareStrings: (a: string, b: string) => number;
export declare function clock(): Trace;
export declare function elapsedMicroseconds(trace: Trace): number;
export declare function internedSpaces(n: number): string;
export declare class RegisterCollection {
    clientCollections: Properties.MapLike<Properties.MapLike<ISegment[]> | undefined>;
    set(clientId: string, id: string, segments: ISegment[]): void;
    get(clientId: string, id: string): ISegment[] | undefined;
    getLength(clientId: string, id: string): number;
    removeClient(clientId: string): void;
}
export interface IConsensusInfo {
    marker: Marker;
    callback: (m: Marker) => void;
}
export interface ClientSeq {
    refSeq: number;
    clientId: string;
}
export declare const clientSeqComparer: Collections.Comparer<ClientSeq>;
export interface LRUSegment {
    segment?: ISegment;
    maxSeq: number;
}
export declare function glc(mergeTree: MergeTree, id: number): string;
export interface SegmentAccumulator {
    segments: ISegment[];
}
export interface MinListener {
    minRequired: number;
    onMinGE(minSeq: number): any;
}
export declare type LocalReferenceMapper = (id: string) => LocalReference;
export declare class MergeTree {
    options?: Properties.PropertySet | undefined;
    static TextSegmentGranularity: number;
    static zamboniSegmentsMaxCount: number;
    static options: {
        incrementalUpdate: boolean;
        insertAfterRemovedSegs: boolean;
        measureOrdinalTime: boolean;
        measureWindowTime: boolean;
        zamboniSegments: boolean;
    };
    static traceAppend: boolean;
    static traceZRemove: boolean;
    static traceOrdinals: boolean;
    static traceGatherText: boolean;
    static diagInsertTie: boolean;
    static skipLeftShift: boolean;
    static diagOverlappingRemove: boolean;
    static traceTraversal: boolean;
    static traceIncrTraversal: boolean;
    static initBlockUpdateActions: BlockUpdateActions;
    static theUnfinishedNode: IMergeBlock;
    static readonly blockUpdateMarkers = true;
    windowTime: number;
    packTime: number;
    ordTime: number;
    maxOrdTime: number;
    root: IMergeBlock;
    blockUpdateActions: BlockUpdateActions;
    collabWindow: CollaborationWindow;
    pendingSegments: Collections.List<SegmentGroup> | undefined;
    segmentsToScour: Collections.Heap<LRUSegment> | undefined;
    idToSegment: Properties.MapLike<ISegment>;
    minSeqListeners: Collections.Heap<MinListener> | undefined;
    getLongClientId?: (id: number) => string;
    mergeTreeDeltaCallback?: MergeTreeDeltaCallback;
    mergeTreeMaintenanceCallback?: MergeTreeMaintenanceCallback;
    constructor(options?: Properties.PropertySet | undefined);
    private makeBlock;
    clone(): void;
    blockClone(block: IMergeBlock, segments?: ISegment[]): MergeBlock;
    private segmentClone;
    localNetLength(segment: ISegment): number;
    mapIdToSegment(id: string, segment: ISegment): void;
    private addNode;
    reloadFromSegments(segments: ISegment[]): void;
    startCollaboration(localClientId: number, minSeq: number, currentSeq: number): void;
    private addToLRUSet;
    private underflow;
    private scourNode;
    private packParent;
    private zamboniSegments;
    getCollabWindow(): CollaborationWindow;
    getStats(): MergeTreeStats;
    findHistorialPosition(pos: number, fromSeq: number, toSeq: number, clientId: number): number;
    findHistorialPositionFromClient(pos: number, fromSeq: number, toSeq: number, clientId: number): number;
    findHistorialRangeFromClient(rangeStart: number, rangeEnd: number, fromSeq: number, toSeq: number, clientId: number): Base.IIntegerRange[];
    findHistorialRange(rangeStart: number, rangeEnd: number, fromSeq: number, toSeq: number, clientId: number): Base.IIntegerRange[];
    getLength(refSeq: number, clientId: number): number;
    /**
     * Returns the current length of the MergeTree for the local client.
     */
    get length(): number;
    getPosition(node: MergeNode, refSeq: number, clientId: number): number;
    cloneSegments(refSeq: number, clientId: number, start?: number, end?: number): ISegment[];
    getContainingSegment<T extends ISegment>(pos: number, refSeq: number, clientId: number): {
        segment: T | undefined;
        offset: number | undefined;
    };
    private blockLength;
    getRemovalInfo(segment: ISegment): IRemovalInfo;
    private nodeLength;
    addMinSeqListener(minRequired: number, onMinGE: (minSeq: number) => void): void;
    private notifyMinSeqListeners;
    setMinSeq(minSeq: number): void;
    referencePositionToLocalPosition(refPos: ReferencePosition, refSeq?: number, clientId?: number): number;
    getStackContext(startPos: number, clientId: number, rangeLabels: string[]): RangeStackMap;
    findTile(startPos: number, clientId: number, tileLabel: string, posPrecedesTile?: boolean): {
        tile: ReferencePosition;
        pos: number;
    } | undefined;
    private search;
    private searchBlock;
    private backwardSearch;
    private backwardSearchBlock;
    private updateRoot;
    /**
     * Assign sequence number to existing segment; update partial lengths to reflect the change
     * @param seq - sequence number given by server to pending segment
     */
    ackPendingSegment(opArgs: IMergeTreeDeltaOpArgs, verboseOps?: boolean): void;
    private addToPendingList;
    getMarkerFromId(id: string): ISegment;
    /**
     * Given a position specified relative to a marker id, lookup the marker
     * and convert the position to a character position.
     * @param relativePos - Id of marker (may be indirect) and whether position is before or after marker.
     * @param refseq - The reference sequence number at which to compute the position.
     * @param clientId - The client id with which to compute the position.
     */
    posFromRelativePos(relativePos: ops.IRelativePosition, refseq?: number, clientId?: number): number;
    insertSegments(pos: number, segments: ISegment[], refSeq: number, clientId: number, seq: number, opArgs: IMergeTreeDeltaOpArgs | undefined): void;
    insertAtReferencePosition(referencePosition: ReferencePosition, insertSegment: ISegment, opArgs: IMergeTreeDeltaOpArgs): void;
    /**
     * Resolves a remote client's position against the local sequence
     * and returns the remote client's position relative to the local
     * sequence
     * @param remoteClientPosition - The remote client's position to resolve
     * @param remoteClientRefSeq - The reference sequence number of the remote client
     * @param remoteClientId - The client id of the remote client
     */
    resolveRemoteClientPosition(remoteClientPosition: number, remoteClientRefSeq: number, remoteClientId: number): number | undefined;
    private insertChildNode;
    private blockInsert;
    private readonly splitLeafSegment;
    private ensureIntervalBoundary;
    private breakTie;
    private leftExcursion;
    private rightExcursion;
    private insertingWalk;
    private split;
    private ordinalIntegrity;
    private nodeOrdinalIntegrity;
    private nodeUpdateOrdinals;
    private addOverlappingClient;
    /**
     * Annotate a range with properties
     * @param start - The inclusive start position of the range to annotate
     * @param end - The exclusive end position of the range to annotate
     * @param props - The properties to annotate the range with
     * @param combiningOp - Optional. Specifies how to combine values for the property, such as "incr" for increment.
     * @param refSeq - The reference sequence number to use to apply the annotate
     * @param clientId - The id of the client making the annotate
     * @param seq - The sequence number of the annotate operation
     * @param opArgs - The op args for the annotate op. this is passed to the merge tree callback if there is one
     */
    annotateRange(start: number, end: number, props: Properties.PropertySet, combiningOp: ops.ICombiningOp | undefined, refSeq: number, clientId: number, seq: number, opArgs: IMergeTreeDeltaOpArgs): void;
    markRangeRemoved(start: number, end: number, refSeq: number, clientId: number, seq: number, overwrite: boolean | undefined, opArgs: IMergeTreeDeltaOpArgs): void;
    private nodeUpdateLengthNewStructure;
    removeLocalReference(segment: ISegment, lref: LocalReference): void;
    addLocalReference(lref: LocalReference): void;
    private blockUpdate;
    private blockUpdatePathLengths;
    private blockUpdateLength;
    map<TClientData>(actions: SegmentActions<TClientData>, refSeq: number, clientId: number, accum: TClientData): void;
    mapRange<TClientData>(actions: SegmentActions<TClientData>, refSeq: number, clientId: number, accum: TClientData, start?: number, end?: number, splitRange?: boolean): void;
    nodeToString(block: IMergeBlock, strbuf: string, indentCount?: number): string;
    toString(): string;
    incrementalBlockMap<TContext>(stateStack: Collections.Stack<IncrementalMapState<TContext>>): void;
    private nodeMap;
    walkAllSegments<TClientData>(block: IMergeBlock, action: (segment: ISegment, accum?: TClientData) => boolean, accum?: TClientData): any;
    private nodeMapReverse;
}
export {};
//# sourceMappingURL=mergeTree.d.ts.map